@@ -1,7 +1,7 @@
-{-# Language CPP, FlexibleContexts, TypeFamilies, KindSignatures, TemplateHaskell, GADTs #-}
+{-# Language CPP, FlexibleContexts, TypeFamilies, KindSignatures, TemplateHaskell, GADTs, RankNTypes, MagicHash, ConstraintKinds, PolyKinds #-}
 
-#if __GLASGOW_HASKELL__ >= 704
-{-# LANGUAGE ConstraintKinds #-}
+#if __GLASGOW_HASKELL__ < 806
+{-# Language TypeInType #-}
 #endif
 
 #if __GLASGOW_HASKELL__ >= 807
@@ -9,8 +9,12 @@
 {-# LANGUAGE TypeApplications #-}
 #endif
 
-#if MIN_VERSION_template_haskell(2,8,0)
-{-# Language PolyKinds #-}
+#if MIN_VERSION_template_haskell(2,21,0)
+{-# Language TypeAbstractions #-}
+#endif
+
+#if MIN_VERSION_template_haskell(2,18,0)
+{-# LANGUAGE UnliftedDatatypes #-}
 #endif
 
 {-|
@@ -27,18 +31,22 @@ the versions of GHC supported by this package.
 -}
 module Main (main) where
 
-#if __GLASGOW_HASKELL__ >= 704
-import           Control.Monad (zipWithM_)
-#endif
-
-import           Control.Monad (unless, when)
+import           Control.Monad (unless, when, zipWithM_)
 import qualified Data.Map as Map
-
-#if MIN_VERSION_base(4,7,0)
+import           Data.Kind
 import           Data.Type.Equality ((:~:)(..))
+
+#if __GLASGOW_HASKELL__ >= 810
+import           GHC.Exts (Any, RuntimeRep(..), TYPE)
+#endif
+#if __GLASGOW_HASKELL__ >= 902
+import           GHC.Exts (UnliftedType, Levity(..))
 #endif
 
-import           Language.Haskell.TH
+import           GHC.Exts (Array#)
+
+import qualified Language.Haskell.TH as TH (Type)
+import           Language.Haskell.TH hiding (Type)
 import           Language.Haskell.TH.Datatype as Datatype
 import           Language.Haskell.TH.Datatype.TyVarBndr
 import           Language.Haskell.TH.Lib (starK)
@@ -59,11 +67,8 @@ main =
      voidstosTest
      strictDemoTest
      recordVanillaTest
-#if MIN_VERSION_template_haskell(2,6,0)
      t43Test
      t58Test
-#endif
-#if MIN_VERSION_template_haskell(2,7,0)
      dataFamilyTest
      ghc78bugTest
      quotedTest
@@ -75,33 +80,22 @@ main =
      t46Test
      t73Test
      t95Test
-#endif
      fixityLookupTest
-#if __GLASGOW_HASKELL__ >= 704
      resolvePredSynonymsTest
-#endif
      reifyDatatypeWithConNameTest
      reifyConstructorTest
-#if MIN_VERSION_base(4,7,0)
      importedEqualityTest
-#endif
-#if MIN_VERSION_template_haskell(2,8,0)
      kindSubstTest
      t59Test
      t61Test
      t66Test
      t80Test
-#endif
-#if MIN_VERSION_template_haskell(2,11,0)
      t79TestA
-#endif
 #if MIN_VERSION_template_haskell(2,19,0)
      t79TestB
 #endif
-#if __GLASGOW_HASKELL__ >= 800
      t37Test
      polyKindedExTyvarTest
-#endif
 #if __GLASGOW_HASKELL__ >= 807
      resolveTypeSynonymsVKATest
 #endif
@@ -113,6 +107,24 @@ main =
 #if MIN_VERSION_template_haskell(2,20,0)
      t100Test
 #endif
+#if MIN_VERSION_template_haskell(2,21,0)
+     t103Test
+#endif
+#if __GLASGOW_HASKELL__ >= 810
+     t107Test
+     t108Test
+#endif
+#if __GLASGOW_HASKELL__ >= 804
+     t110Test
+#endif
+#if MIN_VERSION_template_haskell(2,16,0)
+     unboxedTupleTest
+#endif
+#if MIN_VERSION_template_haskell(2,18,0)
+     unliftedGADTDecTest
+#endif
+     primTyConTest
+
 
 adt1Test :: IO ()
 adt1Test =
@@ -130,6 +142,7 @@ adt1Test =
            , datatypeVars = [aTvb,bTvb]
            , datatypeInstTypes = [aSig, bSig]
            , datatypeVariant = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons =
                [ ConstructorInfo
                    { constructorName = 'Adtc1
@@ -163,6 +176,7 @@ gadt1Test =
            , datatypeVars = [kindedTV a starK]
            , datatypeInstTypes = [SigT aVar starK]
            , datatypeVariant = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons =
                [ ConstructorInfo
                    { constructorName = 'Gadtc1
@@ -210,6 +224,7 @@ gadtrec1Test =
            , datatypeVars      = [kindedTV a starK]
            , datatypeInstTypes = [SigT (VarT a) starK]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ con, con { constructorName = 'Gadtrecc2 } ]
            }
@@ -231,6 +246,7 @@ equalTest =
            , datatypeVars      = [aTvb, bTvb, cTvb]
            , datatypeInstTypes = [aSig, bSig, cSig]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'Equalc
@@ -263,6 +279,7 @@ showableTest =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'Showable
@@ -285,6 +302,7 @@ recordTest =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'R1
@@ -320,6 +338,7 @@ gadt2Test =
            , datatypeVars      = [aTvb, bTvb]
            , datatypeInstTypes = [aSig, bSig]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ con { constructorName = 'Gadt2c1
                      , constructorContext = [equalPred bVar (AppT ListT aVar)] }
@@ -344,6 +363,7 @@ voidstosTest =
            , datatypeVars      = [kindedTV g (arrowKCompat starK starK)]
            , datatypeInstTypes = [SigT (VarT g) (arrowKCompat starK starK)]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      = []
            }
   )
@@ -358,6 +378,7 @@ strictDemoTest =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'StrictDemo
@@ -377,7 +398,6 @@ recordVanillaTest =
   $(do info <- reifyRecord 'gadtrec1a
        validateCI info gadtRecVanillaCI)
 
-#if MIN_VERSION_template_haskell(2,6,0)
 t43Test :: IO ()
 t43Test =
   $(do [decPlain] <- [d| data T43Plain where MkT43Plain :: T43Plain |]
@@ -389,6 +409,7 @@ t43Test =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "MkT43Plain"
@@ -408,6 +429,7 @@ t43Test =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "MkT43Fam"
@@ -432,6 +454,7 @@ t58Test =
            , datatypeVars      = []
            , datatypeInstTypes = []
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "MkFoo"
@@ -442,9 +465,7 @@ t58Test =
                    , constructorVariant    = NormalConstructor } ]
            }
    )
-#endif
 
-#if MIN_VERSION_template_haskell(2,7,0)
 dataFamilyTest :: IO ()
 dataFamilyTest =
   $(do info <- reifyDatatype 'DFMaybe
@@ -456,6 +477,7 @@ dataFamilyTest =
            , datatypeVars      = [kindedTV a starK]
            , datatypeInstTypes = [AppT (ConT ''Maybe) (VarT a)]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'DFMaybe
@@ -479,6 +501,7 @@ ghc78bugTest =
            , datatypeVars      = [kindedTV c starK]
            , datatypeInstTypes = [SigT cVar starK]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'DF1
@@ -503,6 +526,7 @@ quotedTest =
            , datatypeVars      = [plainTV a]
            , datatypeInstTypes = [aVar]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "MkQuoted"
@@ -518,18 +542,15 @@ polyTest :: IO ()
 polyTest =
   $(do info <- reifyDatatype 'MkPoly
        let [a,k] = map mkName ["a","k"]
-           kVar  = varKCompat k
+           kVar  = VarT k
        validateDI info
          DatatypeInfo
            { datatypeName      = ''Poly
            , datatypeContext   = []
-           , datatypeVars      = [
-#if __GLASGOW_HASKELL__ >= 800
-                                 kindedTV k starK,
-#endif
-                                 kindedTV a kVar ]
+           , datatypeVars      = [kindedTV k starK, kindedTV a kVar]
            , datatypeInstTypes = [SigT (VarT a) kVar]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkPoly
@@ -555,6 +576,7 @@ gadtFamTest =
            , datatypeVars      = [cTvb,dTvb]
            , datatypeInstTypes = [cSig,dSig]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkGadtFam1
@@ -613,6 +635,7 @@ famLocalDecTest1 =
            , datatypeVars      = []
            , datatypeInstTypes = [ConT ''Int]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "FamLocalDec1Int"
@@ -638,6 +661,7 @@ famLocalDecTest2 =
            , datatypeVars      = [aTvb,bTvb]
            , datatypeInstTypes = [ConT ''Int, TupleT 2 `AppT` aVar `AppT` bVar, aVar]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "FamLocalDec2Int"
@@ -676,6 +700,7 @@ t73Test =
            , datatypeVars      = [bTvb]
            , datatypeInstTypes = [ConT ''Int, SigT bVar starK]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT73
@@ -700,6 +725,7 @@ t95Test =
            , datatypeVars      = [aTvb]
            , datatypeInstTypes = [AppT ListT aVar]
            , datatypeVariant   = DataInstance
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT95
@@ -710,14 +736,12 @@ t95Test =
                    , constructorVariant    = NormalConstructor }]
            }
    )
-#endif
 
 fixityLookupTest :: IO ()
 fixityLookupTest =
   $(do Just (Fixity 6 InfixR) <- reifyFixityCompat '(:**:)
        [| return () |])
 
-#if __GLASGOW_HASKELL__ >= 704
 resolvePredSynonymsTest :: IO ()
 resolvePredSynonymsTest =
   $(do info <- reifyDatatype ''PredSynT
@@ -729,7 +753,6 @@ resolvePredSynonymsTest =
            test3 = mkTest cxt3 [equalPred (ConT ''Int) (ConT ''Int)]
        mapM_ (either fail return) [test1,test2,test3]
        [| return () |])
-#endif
 
 reifyDatatypeWithConNameTest :: IO ()
 reifyDatatypeWithConNameTest =
@@ -742,6 +765,7 @@ reifyDatatypeWithConNameTest =
           , datatypeVars      = [kindedTV a starK]
           , datatypeInstTypes = [SigT (VarT a) starK]
           , datatypeVariant   = Datatype
+          , datatypeReturnKind = starK
           , datatypeCons      =
               [ ConstructorInfo
                   { constructorName       = 'Nothing
@@ -761,25 +785,24 @@ reifyConstructorTest =
   $(do info <- reifyConstructor 'Just
        validateCI info justCI)
 
-#if MIN_VERSION_base(4,7,0)
 importedEqualityTest :: IO ()
 importedEqualityTest =
   $(do info <- reifyDatatype ''(:~:)
        let names@[a,b] = map mkName ["a","b"]
            [aVar,bVar] = map VarT names
            k           = mkName "k"
-           kKind       = varKCompat k
+           kKind       = VarT k
        validateDI info
          DatatypeInfo
            { datatypeContext   = []
            , datatypeName      = ''(:~:)
-           , datatypeVars      = [
-#if __GLASGOW_HASKELL__ >= 800
-                                 kindedTV k starK,
-#endif
-                                 kindedTV a kKind, kindedTV b kKind]
+           , datatypeVars      = [ kindedTV k starK
+                                 , kindedTV a kKind
+                                 , kindedTV b kKind
+                                 ]
            , datatypeInstTypes = [SigT aVar kKind, SigT bVar kKind]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'Refl
@@ -790,9 +813,7 @@ importedEqualityTest =
                    , constructorVariant    = NormalConstructor } ]
            }
    )
-#endif
 
-#if MIN_VERSION_template_haskell(2,8,0)
 kindSubstTest :: IO ()
 kindSubstTest =
   $(do k1 <- newName "k1"
@@ -801,7 +822,7 @@ kindSubstTest =
        let ty = ForallT [kindedTVSpecified a (VarT k1)] [] (VarT a)
            substTy = applySubstitution (Map.singleton k1 (VarT k2)) ty
 
-           checkFreeVars :: Type -> [Name] -> Q ()
+           checkFreeVars :: TH.Type -> [Name] -> Q ()
            checkFreeVars t freeVars =
              unless (freeVariables t == freeVars) $
                fail $ "free variables of " ++ show t ++ " should be " ++ show freeVars
@@ -817,11 +838,7 @@ t59Test =
        let proxyAK  = ConT (mkName "Proxy") `AppT` SigT (VarT a) (VarT k)
                         -- Proxy (a :: k)
            expected = ForallT
-#if __GLASGOW_HASKELL__ >= 800
                         [plainTVSpecified k, kindedTVSpecified a (VarT k)]
-#else
-                        [kindedTVSpecified a (VarT k)]
-#endif
                         [] proxyAK
            actual = quantifyType proxyAK
        unless (expected == actual) $
@@ -833,7 +850,7 @@ t59Test =
 
 t61Test :: IO ()
 t61Test =
-  $(do let test :: Type -> Type -> Q ()
+  $(do let test :: TH.Type -> TH.Type -> Q ()
            test orig expected = do
              actual <- resolveTypeSynonyms orig
              unless (expected == actual) $
@@ -844,22 +861,18 @@ t61Test =
 
            idAppT = (ConT ''Id `AppT`)
            a = mkName "a"
-       test (SigT (idAppT $ ConT ''Int) (idAppT StarT))
-            (SigT (ConT ''Int) StarT)
-#if MIN_VERSION_template_haskell(2,10,0)
-       test (ForallT [kindedTVSpecified a (idAppT StarT)]
+       test (SigT (idAppT $ ConT ''Int) (idAppT starK))
+            (SigT (ConT ''Int) starK)
+       test (ForallT [kindedTVSpecified a (idAppT starK)]
                      [idAppT (ConT ''Show `AppT` VarT a)]
                      (idAppT $ VarT a))
-            (ForallT [kindedTVSpecified a StarT]
+            (ForallT [kindedTVSpecified a starK]
                      [ConT ''Show `AppT` VarT a]
                      (VarT a))
-#endif
-#if MIN_VERSION_template_haskell(2,11,0)
        test (InfixT (idAppT $ ConT ''Int) ''Either (idAppT $ ConT ''Int))
             (InfixT (ConT ''Int) ''Either (ConT ''Int))
        test (ParensT (idAppT $ ConT ''Int))
             (ConT ''Int)
-#endif
 #if MIN_VERSION_template_haskell(2,19,0)
        test (PromotedInfixT (idAppT $ ConT ''Int) '(:^:) (idAppT $ ConT ''Int))
             (PromotedInfixT (ConT ''Int) '(:^:) (ConT ''Int))
@@ -882,6 +895,7 @@ t66Test =
            , datatypeInstTypes = [ VarT a, VarT b
                                  , SigT (VarT f) fKind, SigT (VarT x) starK ]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = mkName "MkFoo"
@@ -912,9 +926,7 @@ t80Test = do
                    , "Actual:   " ++ pprint actual
                    ]
   return ()
-#endif
 
-#if MIN_VERSION_template_haskell(2,11,0)
 t79TestA :: IO ()
 t79TestA =
   $(do let [a,b,c]  = map mkName ["a","b","c"]
@@ -930,7 +942,6 @@ t79TestA =
                         , "Actual:   " ++ pprint actual
                         ]
        [| return () |])
-#endif
 
 #if MIN_VERSION_template_haskell(2,19,0)
 t79TestB :: IO ()
@@ -950,7 +961,6 @@ t79TestB =
        [| return () |])
 #endif
 
-#if __GLASGOW_HASKELL__ >= 800
 t37Test :: IO ()
 t37Test =
   $(do infoA <- reifyDatatype ''T37a
@@ -967,6 +977,7 @@ t37Test =
            , datatypeVars      = [kTvb, aTvb]
            , datatypeInstTypes = [kSig, aSig]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT37a
@@ -985,6 +996,7 @@ t37Test =
            , datatypeVars      = [kTvb, aTvb]
            , datatypeInstTypes = [aSig]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT37b
@@ -1003,6 +1015,7 @@ t37Test =
            , datatypeVars      = [kTvb, aTvb]
            , datatypeInstTypes = [aSig]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT37c
@@ -1026,6 +1039,7 @@ polyKindedExTyvarTest =
            , datatypeVars      = [kindedTV a starK]
            , datatypeInstTypes = [SigT aVar starK]
            , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
            , datatypeCons      =
                [ ConstructorInfo
                    { constructorName       = 'MkT48
@@ -1067,7 +1081,6 @@ t75Test =
                     ++ show (length cs)
        [| return () |]
    )
-#endif
 
 #if __GLASGOW_HASKELL__ >= 807
 resolveTypeSynonymsVKATest :: IO ()
@@ -1161,6 +1174,7 @@ t100Test =
                , datatypeVars = []
                , datatypeInstTypes = []
                , datatypeVariant = Datatype.TypeData
+               , datatypeReturnKind = starK
                , datatypeCons =
                    [ ConstructorInfo
                        { constructorName = ''MkT100
@@ -1179,3 +1193,215 @@ t100Test =
        validateDI mkT100Info expectedInfo
    )
 #endif
+
+#if MIN_VERSION_template_haskell(2,21,0)
+t103Test :: IO ()
+t103Test =
+  $(do [dec] <- [d| data T102 @k (a :: k) |]
+       info <- normalizeDec dec
+       let k = mkName "k"
+           a = mkName "a"
+       validateDI info
+         DatatypeInfo
+           { datatypeName      = mkName "T102"
+           , datatypeContext   = []
+           , datatypeVars      = [plainTV k, kindedTV a (VarT k)]
+           , datatypeInstTypes = [SigT (VarT a) (VarT k)]
+           , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
+           , datatypeCons      = []
+           }
+   )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 810
+t107Test :: IO ()
+t107Test =
+  $(do info <- reifyDatatype ''T107
+       let r = mkName "r"
+       validateDI info
+         DatatypeInfo
+           { datatypeName      = mkName "T107"
+           , datatypeContext   = []
+           , datatypeVars      = [kindedTV r (ConT ''RuntimeRep)]
+           , datatypeInstTypes = []
+           , datatypeVariant   = Newtype
+           , datatypeReturnKind = ConT ''TYPE `AppT` VarT r
+           , datatypeCons      =
+               [ ConstructorInfo
+                   { constructorName       = mkName "MkT107"
+                   , constructorVars       = []
+                   , constructorContext    = []
+                   , constructorFields     = [ConT ''Any `SigT` (ConT ''TYPE `AppT` VarT r)]
+                   , constructorStrictness = [notStrictAnnot]
+                   , constructorVariant    = NormalConstructor
+                   }
+               ]
+           }
+   )
+
+t108Test :: IO ()
+t108Test =
+  $(do [dec] <- [d| data T108 :: forall k -> k -> Type where
+                      MkT108 :: forall k (a :: k). T108 k a
+                  |]
+       info <- normalizeDec dec
+       let k = mkName "k"
+           a = mkName "a"
+       validateDI info
+         DatatypeInfo
+           { datatypeName      = mkName "T108"
+           , datatypeContext   = []
+           , datatypeVars      = [plainTV k, kindedTV a (VarT k)]
+           , datatypeInstTypes = [VarT k, SigT (VarT a) (VarT k)]
+           , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
+           , datatypeCons      =
+               [ ConstructorInfo
+                   { constructorName       = mkName "MkT108"
+                   , constructorVars       = []
+                   , constructorContext    = []
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor
+                   }
+               ]
+           }
+   )
+#endif
+
+#if __GLASGOW_HASKELL__ >= 804
+t110Test :: IO ()
+t110Test =
+  $(do [dec] <- [d| data T110 :: forall k. k -> Type where
+                      MkT110 :: forall k (a :: k). T110 a
+                  |]
+       info <- normalizeDec dec
+       let k = mkName "k"
+           a = mkName "a"
+       validateDI info
+         DatatypeInfo
+           { datatypeName      = mkName "T110"
+           , datatypeContext   = []
+           , datatypeVars      = [plainTV k, kindedTV a (VarT k)]
+           , datatypeInstTypes = [SigT (VarT a) (VarT k)]
+           , datatypeVariant   = Datatype
+           , datatypeReturnKind = starK
+           , datatypeCons      =
+               [ ConstructorInfo
+                   { constructorName       = mkName "MkT110"
+                   , constructorVars       = []
+                   , constructorContext    = []
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor
+                   }
+               ]
+           }
+   )
+#endif
+
+#if MIN_VERSION_template_haskell(2,16,0)
+unboxedTupleTest :: IO ()
+unboxedTupleTest =
+  $(do k0 <- newName "k0"
+       k1 <- newName "k1"
+       a <- newName "a"
+       b  <- newName "b"
+       tupleInfo <- reifyDatatype (unboxedTupleTypeName 2)
+       validateDI tupleInfo
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName = unboxedTupleTypeName 2
+           , datatypeVars = [kindedTV k0 starK
+                            ,kindedTV a (AppT (ConT ''TYPE) (VarT k0 ))
+                            ,kindedTV k1 starK
+                            ,kindedTV b (AppT (ConT ''TYPE) (VarT k1))]
+           , datatypeInstTypes = [SigT (VarT a) (AppT (ConT ''TYPE) (VarT k0))
+                                 ,SigT (VarT b) (AppT (ConT ''TYPE) (VarT k1))]
+           , datatypeVariant = Datatype
+           , datatypeReturnKind =
+               AppT
+                 (ConT ''TYPE)
+                 (AppT
+                    (PromotedT 'TupleRep)
+                    (AppT
+                      (AppT PromotedConsT (VarT k0))
+                        (AppT
+                          (AppT PromotedConsT (VarT k1))
+                          (SigT PromotedNilT (AppT ListT (ConT ''RuntimeRep))))))
+           , datatypeCons =
+             [ ConstructorInfo
+               { constructorName = unboxedTupleDataName 2
+               , constructorVars = []
+               , constructorContext = []
+               , constructorFields = [VarT a, VarT b]
+               , constructorStrictness = [notStrictAnnot, notStrictAnnot]
+               , constructorVariant = NormalConstructor}]
+          }
+  )
+#endif
+
+#if MIN_VERSION_template_haskell(2,18,0)
+unliftedGADTDecTest :: IO ()
+unliftedGADTDecTest =
+  $(do a <- newName "a"
+       s <- newName "s"
+       [dec] <- [d| data UnliftedGADT a :: UnliftedType where
+                      UnliftedGADT :: Show s => s -> a -> UnliftedGADT a
+                |]
+       info <- normalizeDec dec
+       validateDI info
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName = mkName "UnliftedGADT"
+           , datatypeVars = [plainTV a]
+           , datatypeInstTypes = [VarT a]
+           , datatypeVariant = Datatype
+           , datatypeReturnKind = ConT ''TYPE `AppT` (PromotedT 'BoxedRep `AppT` PromotedT 'Unlifted)
+           , datatypeCons =
+               [ConstructorInfo
+                  {constructorName = mkName "UnliftedGADT"
+                  , constructorVars = [plainTV s]
+                  , constructorContext = [AppT (ConT ''Show) (VarT s)]
+                  , constructorFields = [VarT s,VarT a]
+                  , constructorStrictness = [notStrictAnnot, notStrictAnnot]
+                  , constructorVariant = NormalConstructor}
+               ]
+           }
+   )
+#endif
+
+
+primTyConTest :: IO ()
+primTyConTest =
+  $(do l <- newName "l"
+       a <- newName "a"
+       info <- reifyDatatype ''Array#
+       validateDI info
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName = mkName "Array#"
+#if MIN_VERSION_template_haskell(2,19,0)
+           , datatypeVars = [kindedTV l (ConT ''Levity)
+                            , kindedTV a (ConT ''TYPE `AppT` (PromotedT 'BoxedRep `AppT` VarT l))
+                            ]
+           , datatypeInstTypes = [SigT (VarT a) (ConT ''TYPE `AppT` (PromotedT 'BoxedRep `AppT` VarT l))]
+           , datatypeReturnKind = ConT ''TYPE `AppT` (PromotedT 'BoxedRep `AppT` PromotedT 'Unlifted)
+#elif MIN_VERSION_template_haskell(2,18,0)
+           , datatypeVars = [ kindedTV a StarT]
+           , datatypeInstTypes = [SigT (VarT a) StarT]
+           , datatypeReturnKind = ConT ''TYPE `AppT` (PromotedT 'BoxedRep `AppT` PromotedT 'Unlifted)
+#elif MIN_VERSION_template_haskell(2,16,0)
+           , datatypeVars = [kindedTV a starK]
+           , datatypeInstTypes = [SigT (VarT a) starK]
+           , datatypeReturnKind = ConT ''TYPE `AppT` PromotedT 'UnliftedRep
+#else
+           , datatypeVars = [kindedTV a starK]
+           , datatypeInstTypes = [SigT (VarT a) starK]
+           , datatypeReturnKind = starK
+#endif
+           , datatypeVariant = Datatype
+           , datatypeCons = []
+           }
+   )
